/*
 * ppm.c --- load/save PPM format file.
 *
 *			by k-chinen@is.aist-nara.ac.jp, 1994
 *
 * $Id: ppm.c,v 1.1 1994/05/12 22:32:01 k-chinen Exp k-chinen $
 */

/*
 * Note: X's pixel is "long". So we need array of "long" for drawing.
 *       This program require size of "long" >= 4 and bits of "long" >= 32. 
 */

#define _DEBUG_PPM

#include "sp.h"

#include <fcntl.h>
#include <limits.h>
#include <ctype.h>

/* PATH, COLOR, LINE */
#define PSIZE 256
#define CSIZE 256
#define LSIZE 1000

int ppm_read(
	char *name,
	char **rbuf, char **gbuf, char **bbuf,
	int *w, int *h, int *cn)
{
	char rname[PSIZE];
	char cmd[LSIZE];
	char *p;
	int fd;
	int c,n;
	int wid,ht,cnum;
	int l,s;
	int k=0;
	char *red, *green, *blue;


	sprintf(rname,"#sp_tmp.%d.ppm", getpid());
	sprintf(cmd,"convert %s \\%s",name,rname);
	system(cmd);
#ifdef DEBUG_PPM
	fprintf(stderr,"ppm_read: OK, I apply\n\t%s\n",cmd);
#endif

	/* read raw file */
	if((fd=open(rname,O_RDONLY,0644))<0) {
		fprintf(stderr,"ppm_read: cannot open \"%s\"\n",rname);
		return 0;
	}


	c = read(fd, cmd, sizeof(cmd));

#ifdef DEBUG_PPM
	fprintf(stderr,"ppm_read: OK, Read %d char from \"%s\"\n",c,rname);
#endif

	n = 0;
	p = cmd;
	while(p-cmd<c && n<2) {
#ifdef DEBUG_PPM
		fprintf(stdout,"# top char of line,%#x = (%d) %c ; ",p,*p,*p);
#endif

		/* skip whites of line head */
		while(p-cmd<c && *p==' ' && *p=='\t')
			p++;

		if(*p=='P') {
#ifdef DEBUG_PPM
			fprintf(stdout,"Skip P-line.\n");
#endif
		}
		else
		if(*p=='#') {
#ifdef DEBUG_PPM
			fprintf(stdout,"Skip #-line.\n");
#endif
		}
		else
		if(isdigit(*p)) {
			/* eveyline have \n */
			if(n==0) {
				sscanf(p, "%d %d", &wid, &ht);
				n++;
#ifdef DEBUG_PPM
				fprintf(stdout,"wid=%d, ht=%d.\n",wid,ht);
#endif
			}
			else
			if(n==1) {
				sscanf(p, "%d", &cnum);
				n++;
#ifdef DEBUG_PPM
				fprintf(stdout,"cnum=%d.\n",cnum);
#endif
			}
			else {
				break;
			}
		}
		else {
			break;
		}

		/* skip rest of lines */
		while(p-cmd<c && *p!='\n')
			p++;
		p++;
	}
	fflush(stderr);
	fflush(stdout);

	if(p-cmd>=c) {
		/*
		fprintf(stderr,"too short file... p=%#p,cmd=%#p,c=%d\n",p,cmd,c);
		*/
		fprintf(stderr,"Ouch! head overrun buffer... p=%#p,cmd=%#p,c=%d\n",p,cmd,c);
		return 0;
	}

	k = 1;

#ifdef DEBUG_PPM
	fprintf(stderr,"# p=%#p,cmd=%#p,c=%d\n",p,cmd,c);
#endif

	c -= p-cmd;
	l = wid*ht;

	red   = (char *)malloc(l*sizeof(char));
	green = (char *)malloc(l*sizeof(char));
	blue  = (char *)malloc(l*sizeof(char));


#ifdef DEBUG_PPM
	fprintf(stderr,"# c=%d\n",c);
#endif
	n = 0;
	s = 0;
	while(c && n<l) {
		if(s==0)
			red[n] = *p;
		if(s==1)
			green[n] = *p;
		if(s==2)
			blue[n] = *p;
		
		p ++;
		c --;
		if(!c) {
			if(!(c=read(fd ,cmd, sizeof(cmd))))
				break;
			p = cmd;
		}
		if(s==2) {
			n++;
			s=0;
		}
		else
			s++;
	}

	close(fd);

	/* Delete temp-file */
	unlink(rname);

#ifdef DEBUG_PPM
	fprintf(stderr,"Ok, I read %dth byte(0-orig) in %d (%d x %d)byte.\n",
		n,l,wid,ht);
#endif
	
	/* return */
	*w = wid;
	*h = ht;
	*cn = cnum;
	*rbuf = red;
	*gbuf = green;
	*bbuf = blue;

	return l;
}

/* 
 * Filter Mask:
 *		if you set to 2, use to 1/4 colors of all colors in ppm-file
 *		if you set to 3, use to 1/8 colors of all colors in ppm-file
 */
#define FLTMASK (5)

char *ppm_conv(
	Display *dpy, Colormap cmap,
	long **cbuf,
	char **rbuf, char **gbuf, char **bbuf,
	int *w, int *h, int *cn,
	int lc)
{
	unsigned long d[CSIZE], oc[CSIZE], t;
	register int i;
	unsigned int rmin, rmax, gmin, gmax, bmin, bmax;
	unsigned int rp, gp, bp;
	unsigned char *red, *green, *blue;
	long *buf;
	int l;
	int n;
	XColor c;
	int ll;
	Status st;
	register unsigned char *p;
	register unsigned char *r, *g, *b;

	red   = (unsigned char*) *rbuf;
	green = (unsigned char*) *gbuf;
	blue  = (unsigned char*) *bbuf;

#ifdef DEBUG_PPM
	printf("ppm_conv: %d x %d\n", *w, *h);
#endif

	l = *w * *h;

	ll = (int)sqrt((double)lc);
	

	buf = (long*)malloc(sizeof(long)*l);
	memset(d, 0x00, sizeof(long)*CSIZE);

	r = red; g = green; b = blue;

	for(i=0;i<l;i++) {
		t = ((*r >> FLTMASK) << 24)
			| ((*g >> FLTMASK) << 16)
			| ((*b >> FLTMASK) << 8)
			| 0xff;

		/* Scan color in Pallete */
		n = 0;
		while(d[n] && t!=d[n] && n<CSIZE) {
			n++;
		}

		/*
		printf("%4d: RGB=(%2x,%2x,%2x) t=%8x, n=%4d, d[n]=%8x\n",
			i, *r, *g, *b, t, n, d[n]);
		*/

		if(t==d[n]) {
			buf[i] = oc[n];
		}
		else
		/* Alloc Color */
		if(!d[n]) {
			c.red   = (short)((*r>>FLTMASK) << (8+FLTMASK));
			c.green = (short)((*g>>FLTMASK) << (8+FLTMASK));
			c.blue  = (short)((*b>>FLTMASK) << (8+FLTMASK));

			/* if Success, st==1 */
			if(st = XAllocColor(dpy, cmap, &c)) {
				oc[n] = buf[i] = c.pixel;
			}
			else {
				oc[n] = buf[i] = 0L;
			}

#ifdef DEBUG_PPM
			printf("%3u: loc=%7u rgb=(%8x,%8x,%8x) col=%3u st=%3u\n",
				n, i, c.red, c.green, c.blue, buf[i], st);
#endif

			d[n] = t;
		}
		else {
			/* 4 is dummy */
			buf[i] = 4L;
		}

		r++; g++; b++;
	}

#ifdef DEBUG_PPM
	printf("ppm_conv: generate %d colors.\n",n);

	n = 0;
	while(d[n] && n<CSIZE) {
		printf("pallet %3d = color %3d\n", n, oc[n]);
		n++;
	}

	printf("ppm_conv: OK, convert done.\n");
#endif 

	*cbuf = buf;
}

#undef FLTMASK


/*
	fname:	filename
	lcols:	limit colors
	ncols:	# of colors
	cols:	array of color
	buf:	buffer for image
 */

/*
int ppm_read(
	char *name,
	char **rbuf, char **gbuf, char **bbuf,
	int *w, int *h, int *cn)
*/
/*
char *ppm_conv(
	Display *dpy, Colormap cmap,
	long **cbuf,
	char **rbuf, char **gbuf, char **bbuf,
	int *w, int *h, int *cn,
	int lc)
*/

int ppm_load(Display *dpy, Colormap cmap,
	char *fname, int lcols, int ncol, long **buf,
	int *wid, int *ht, int cn)
{
	char *red, *green, *blue;

	/* 	Read phisical file and convert it to raw data,
		therefor it was maybe compressed. */
	if(ppm_read(fname,
		&red, &green, &blue,
		wid, ht, &cn)) {

#ifdef DEBUG_PPM
		printf("ppm_load: Read %d x %d, try to convert to pixels\n",*wid,*ht);
#endif

		ppm_conv(dpy, cmap,
			buf,
			&red, &green, &blue,
			wid, ht, &cn,
			lcols);
		
		free(red);
		free(green);
		free(blue);

		return 0;
	}
	else {
		return 1;
	}
}

